add_to_path = "/Users/tariq/Google Drive/Programming Projects/BioVida"
training_data_path = "/Users/tariq/Desktop/biovida_training/actual_training_images"
biovida_path = "/Users/tariq/Google Drive/Programming Projects/BioVida"